import argparse
import json
import logging
import os
import re
import sys
import textwrap
from pprint import pprint
from typing import Iterable, Optional, Set, Tuple

import colorama
import jsonpatch
import square.k8s as k8s
import square.manio as manio
import yaml
from square import __version__
from square.dtypes import (
    RESOURCE_ALIASES, SUPPORTED_KINDS, Config, DeltaCreate, DeltaDelete,
    DeltaPatch, DeploymentPlan, Filepath, JsonPatch, MetaManifest, Selectors,
    ServerManifests,
)

# Convenience: global logger instance to avoid repetitive code.
logit = logging.getLogger("square")


def parse_commandline_args():
    """Return parsed command line."""
    name = os.path.basename(__file__)
    description = textwrap.dedent(f'''
    Manage Kubernetes manifests.

    Examples:
      {name} get
      {name} plan
      {name} apply
    ''')

    def _validate_kind(kind: str) -> str:
        """Convert resource `kind` from aliases to canonical name.
        For instance, `svc` -> `Service`.
        """
        kind = kind.lower()

        # The "all" resource is special - do not expand.
        if kind == "all":
            return "all"

        out = [
            canonical for canonical, aliases in RESOURCE_ALIASES.items()
            if kind in aliases
        ]

        # Must have found at most one or there is a serious bug.
        if len(out) != 1:
            raise argparse.ArgumentTypeError(kind)
        return out[0]

    def _validate_label(label: str) -> Tuple[str, ...]:
        """Convert resource `kind` from aliases to canonical name.
        For instance, `svc` -> `Service`.
        """
        pat = re.compile(r"^[a-z0-9][-a-z0-9_.]*=[-A-Za-z0-9_.]*[A-Za-z0-9]$")
        if pat.match(label) is None:
            raise argparse.ArgumentTypeError(label)
        return tuple(label.split("="))

    # A dummy top level parser that will become the parent for all sub-parsers
    # to share all its arguments.
    parent = argparse.ArgumentParser(
        description=description,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        add_help=False
    )
    parent.add_argument(
        "-v", "--verbosity", action="count", default=0,
        help="Specify multiple times to increase log level."
             " -v: WARNING -vv: INFO -vvv: DEBUG"
    )
    parent.add_argument(
        "-n", type=str, nargs="*",
        metavar="ns", dest="namespaces",
        help="List of namespaces (omit to consider all)",
    )
    parent.add_argument(
        "-l", type=_validate_label, nargs="*",
        metavar="labels", dest="labels", default=tuple(),
        help="List of K8s resources to consider or just 'all'",
    )
    parent.add_argument(
        "--kubeconfig", type=str, metavar="path",
        default=os.environ.get("KUBECONFIG", None),
        help="Location of kubeconfig file (defaults to env KUBECONFIG)",
    )
    parent.add_argument(
        "--folder", type=str, metavar="path",
        default=os.environ.get("SQUARE_FOLDER", "./"),
        help="Manifest folder (defaults to env SQUARE_FOLDER)",
    )
    parent.add_argument(
        "--context", type=str, metavar="ctx", dest="ctx", default=None,
        help="Kubernetes context (use default one if unspecified)",
    )

    # The primary parser for the top level options (eg GET, PATCH, ...).
    parser = argparse.ArgumentParser(add_help=True)
    subparsers = parser.add_subparsers(
        help='Mode', dest='parser', metavar="ACTION",
        title="Operation", required=True
    )

    # Configuration for `kinds` positional arguments. Every sub-parser must
    # specify this one individually and here we define the kwargs to reduce
    # duplicate code.
    kinds_kwargs = {
        "dest": "kinds",
        "type": _validate_kind,
        "nargs": '+',
        "metavar": "resource",
    }

    # Sub-command GET.
    parser_get = subparsers.add_parser(
        'get', help="Get manifests from K8s and save them locally", parents=[parent]
    )
    parser_get.add_argument(**kinds_kwargs)

    # Sub-command DIFF.
    parser_plan = subparsers.add_parser(
        'plan', help="Diff local and K8s manifests", parents=[parent]
    )
    parser_plan.add_argument(**kinds_kwargs)

    # Sub-command PATCH.
    parser_apply = subparsers.add_parser(
        'apply', help="Patch K8s to match local manifests", parents=[parent]
    )
    parser_apply.add_argument(**kinds_kwargs)

    # Sub-command VERSION.
    subparsers.add_parser(
        'version', help="Show Square version and exit", parents=[parent]
    )

    # Parse the actual arguments.
    param = parser.parse_args()

    # NOTE: skip this step if the user for `version` requests since
    # the relevant variables will not be be defined then.
    if param.parser == "version":
        return param

    if not param.kubeconfig:
        print("ERROR: must either specify --kubeconfig or set KUBECONFIG")
        sys.exit(1)

    # Remove duplicates but retain the original order of "param.kinds". This is
    # a "trick" that will only work in Python 3.7+ because it guarantees a
    # stable insertion order for dicts (but not sets).
    param.kinds = list(dict.fromkeys(param.kinds))

    # Expand the "all" resource (if present).
    if "all" in param.kinds:
        param.kinds = list(SUPPORTED_KINDS)

    # Make label list immutable.
    param.labels = tuple(param.labels)

    return param


def make_patch(
        config: Config,
        local: ServerManifests,
        server: ServerManifests) -> Tuple[Optional[JsonPatch], bool]:
    """Return JSON patch to transition `server` to `local`.

    Inputs:
        local: LocalManifests
            Usually the dictionary keys returned by `load_manifest`.
        server: ServerManifests
            Usually the dictionary keys returned by `manio.download`.

    Returns:
        Patch: the JSON patch and human readable diff in a `Patch` tuple.

    """
    # Reduce local and server manifests to salient fields (ie apiVersion, kind,
    # metadata and spec). Abort on error.
    loc, err1 = manio.strip(config, local)
    srv, err2 = manio.strip(config, server)
    if err1 or err2 or loc is None or srv is None:
        return (None, True)

    # Log the manifest info for which we will try to compute a patch.
    man_id = f"{loc.kind.upper()}: {loc.metadata.name}/{loc.metadata.name}"
    logit.debug(f"Making patch for {man_id}")

    # Sanity checks: abort if the manifests do not specify the same resource.
    try:
        assert srv.apiVersion == loc.apiVersion
        assert srv.kind == loc.kind
        assert srv.metadata.name == loc.metadata.name

        # Not all resources live in a namespace, for instance Namespaces,
        # ClusterRoles, ClusterRoleBindings. Here we ensure that the namespace
        # in the local and server manifest matches for those resources that
        # have a namespace.
        if srv.kind in {"Namespace", "ClusterRole", "ClusterRoleBinding"}:
            namespace = None
        else:
            assert srv.metadata.namespace == loc.metadata.namespace
            namespace = srv.metadata.namespace
    except AssertionError:
        # Log the invalid manifests and return with an error.
        keys = ("apiVersion", "kind", "metadata")
        loc_tmp = {k: loc[k] for k in keys}
        srv_tmp = {k: srv[k] for k in keys}
        logit.error(
            "Cannot compute JSON patch for incompatible manifests. "
            f"Local: <{loc_tmp}>  Server: <{srv_tmp}>"
        )
        return (None, True)

    # Determine the PATCH URL for the resource.
    url, err = k8s.urlpath(config, srv.kind, namespace)
    if err:
        return (None, True)
    full_url = f'{url}/{srv.metadata.name}'

    # Compute JSON patch.
    patch = jsonpatch.make_patch(srv, loc)
    patch = json.loads(patch.to_string())

    # Return the patch.
    return (JsonPatch(full_url, patch), False)


def partition_manifests(
        local: ServerManifests,
        server: ServerManifests) -> Tuple[Optional[DeploymentPlan], bool]:
    """Compile `{local,server}` into CREATE, PATCH and DELETE groups.

    The returned deployment plan will contain *every* resource in
    `local` and `server` *exactly once*. Their relative
    order will also be preserved.

    Create: all resources that exist in `local` but not in `server`.
    Delete: all resources that exist in `server` but not in `local`.
    Patch : all resources that exist in both and therefore *may* need patching.

    Inputs:
        local: Dict[MetaManifest:str]
            Usually the dictionary keys returned by `load_manifest`.
        server: Dict[MetaManifest:str]
            Usually the dictionary keys returned by `manio.download`.

    Returns:
        DeploymentPlan

    """
    # Determine what needs adding, removing and patching to steer the K8s setup
    # towards what `local` specifies.
    loc = set(local.keys())
    srv = set(server.keys())
    create = loc - srv
    patch = loc.intersection(srv)
    delete = srv - loc

    # Convert the sets to list. Preserve the relative element ordering as it
    # was in `{local_server}`.
    create_l = [_ for _ in local if _ in create]
    patch_l = [_ for _ in local if _ in patch]
    delete_l = [_ for _ in server if _ in delete]

    # Return the deployment plan.
    plan = DeploymentPlan(create_l, patch_l, delete_l)
    return (plan, False)


def compile_plan(
        config: Config,
        local: ServerManifests,
        server: ServerManifests) -> Tuple[Optional[DeploymentPlan], bool]:
    """Return the `DeploymentPlan` to transition K8s to state of `local`.

    The deployment plan is a named tuple. It specifies which resources to
    create, patch and delete to ensure that the state of K8s matches that
    specified in `local`.

    Inputs:
        config: Config
        local: ServerManifests
            Should be output from `load_manifest` or `load`.
        server: ServerManifests
            Should be output from `manio.download`.

    Returns:
        DeploymentPlan

    """
    # Partition the set of meta manifests into create/delete/patch groups.
    plan, err = partition_manifests(local, server)
    if err or not plan:
        return (None, True)

    # Sanity check: the resources to patch *must* exist in both local and
    # server manifests. This is a bug if not.
    assert set(plan.patch).issubset(set(local.keys()))
    assert set(plan.patch).issubset(set(server.keys()))

    # Compile the Deltas to create the missing resources.
    create = []
    for delta in plan.create:
        url, err = k8s.urlpath(config, delta.kind, namespace=delta.namespace)
        if err:
            return (None, True)
        create.append(DeltaCreate(delta, url, local[delta]))

    # Compile the Deltas to delete the excess resources. Every DELETE request
    # will have to pass along a `DeleteOptions` manifest (see below).
    del_opts = {
        "apiVersion": "v1",
        "kind": "DeleteOptions",
        "gracePeriodSeconds": 0,
        "orphanDependents": False,
    }
    delete = []
    for meta in plan.delete:
        # Resource URL.
        url, err = k8s.urlpath(config, meta.kind, namespace=meta.namespace)
        if err:
            return (None, True)

        # DELETE requests must specify the resource name in the path.
        url = f"{url}/{meta.name}"

        # Assemble the meta and add it to the list.
        delete.append(DeltaDelete(meta, url, del_opts.copy()))

    # Iterate over each manifest that needs patching and determine the
    # necessary JSON Patch to transition K8s into the state specified in the
    # local manifests.
    patches = []
    for meta in plan.patch:
        # Compute textual diff (only useful for the user to study the diff).
        diff_str, err = manio.diff(config, local[meta], server[meta])
        if err:
            return (None, True)

        # Compute the JSON patch that will match K8s to the local manifest.
        patch, err = make_patch(config, local[meta], server[meta])
        if err:
            return (None, True)
        patches.append(DeltaPatch(meta, diff_str, patch))

    # Assemble and return the deployment plan.
    return (DeploymentPlan(create, patches, delete), False)


def print_deltas(plan: DeploymentPlan) -> Tuple[None, bool]:
    """Print human readable version of `plan` to terminal.

    Inputs:
        plan: DeploymentPlan

    Returns:
        None

    """
    # Terminal colours for convenience.
    cAdd = colorama.Fore.GREEN
    cDel = colorama.Fore.RED
    cMod = colorama.Fore.YELLOW
    cReset = colorama.Fore.RESET

    n_add, n_mod, n_del = 0, 0, 0

    # Use Green to list all the resources that we should create.
    for delta in plan.create:
        name = f"{delta.meta.kind.upper()} {delta.meta.namespace}/{delta.meta.name}"

        # Convert manifest to YAML string and print every line in Green.
        txt = yaml.dump(delta.manifest, default_flow_style=False)
        txt = [cAdd + line + cReset for line in txt.splitlines()]

        # Add header line.
        txt = [f"    {line}" for line in txt]
        txt.insert(0, cAdd + f"Create {name}" + cReset)

        # Print the reassembled string.
        print(str.join('\n', txt) + '\n')
        n_add += 1

    # Print the diff (already contains terminal colours) for all the resources
    # that we should patch.
    for delta in plan.patch:
        if len(delta.diff) == 0:
            continue

        # Add some terminal colours to make it look prettier.
        colour_lines = []
        for line in delta.diff.splitlines():
            if line.startswith('+'):
                colour_lines.append(cAdd + line + cReset)
            elif line.startswith('-'):
                colour_lines.append(cDel + line + cReset)
            else:
                colour_lines.append(line)
        colour_lines = [f"    {line}" for line in colour_lines]
        formatted_diff = str.join('\n', colour_lines)

        name = f"{delta.meta.kind.upper()} {delta.meta.namespace}/{delta.meta.name}"
        print(cMod + f"Patch {name}" + cReset + "\n" + formatted_diff + "\n")
        n_mod += 1

    # Use Red to list all the resources that we should delete.
    for delta in plan.delete:
        name = f"{delta.meta.kind.upper()} {delta.meta.namespace}/{delta.meta.name}"
        print(cDel + f"Delete {name}" + cReset)
        n_del += 1

    print("-" * 80)
    print("Plan: " +                         # noqa
          cAdd + f"{n_add:,} to add, " +     # noqa
          cMod + f"{n_mod:,} to change, " +  # noqa
          cDel + f"{n_del:,} to destroy." +  # noqa
          cReset + "\n")
    return (None, False)


def find_namespace_orphans(
        meta_manifests: Iterable[MetaManifest]
) -> Tuple[Optional[Set[MetaManifest]], bool]:
    """Return all orphaned resources in the `meta_manifest` set.

    A resource is orphaned iff it lives in a namespace that is not explicitly
    declared in the set of `meta_manifests`.

    This function is particularly useful to verify a set of local manifests and
    pinpoint resources that someone forgot to delete and that have a typo in
    their namespace field.

    Inputs:
        meta_manifests: Iterable[MetaManifest]

    Returns:
        set[MetaManifest]: orphaned resources.

    """
    # Turn the input into a set.
    meta_manifests = set(meta_manifests)

    # Extract all declared namespaces so we can find the orphans next.
    namespaces = {_.name for _ in meta_manifests if _.kind == 'Namespace'}

    # Find all manifests that are neither a Namespace nor belong to any of the
    # `namespaces` from the previous step
    orphans = {
        _ for _ in meta_manifests
        if _.kind != "Namespace" and _.namespace not in namespaces
    }

    # Return the result.
    return (orphans, True)


def setup_logging(log_level: int) -> None:
    """Configure logging at `log_level`.

    Level 0: ERROR
    Level 1: WARNING
    Level 2: INFO
    Level >=3: DEBUG

    Inputs:
        log_level: int

    Returns:
        None

    """
    # Pick the correct log level.
    if log_level == 0:
        level = "ERROR"
    elif log_level == 1:
        level = "WARNING"
    elif log_level == 2:
        level = "INFO"
    else:
        level = "DEBUG"

    # Create logger.
    logger = logging.getLogger("square")
    logger.setLevel(level)

    # Configure stdout handler.
    ch = logging.StreamHandler()
    ch.setLevel(level)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - "
        f"%(filename)s:%(funcName)s:%(lineno)d - "
        f"%(message)s"
    )
    ch.setFormatter(formatter)

    # Attach stdout handlers to the `square` logger.
    logger.addHandler(ch)
    logit.info(f"Set log level to {level}")


def user_confirmed(answer: str = "yes") -> bool:
    """Return True iff the user answers with `answer`."""
    assert answer, "BUG: desired answer must be non-empty string"
    print(f"Apply the changes?")
    print(f'Only "{answer}" will commence the rollout.')

    try:
        return input("  Your answer: ") == answer
    except KeyboardInterrupt:
        print()
        return False


def main_apply(
        config: Config,
        client,
        folder: Filepath,
        selectors: Selectors,
) -> Tuple[None, bool]:
    """Update K8s to match the specifications in `local_manifests`.

    Create a deployment plan that will transition the K8s state
    `server_manifests` to the desired `local_manifests`.

    Inputs:
        config: Config
        client: `requests` session with correct K8s certificates.
        folder: Filepath
            Path to local manifests eg "./foo"
        selectors: Selectors
            Only operate on resources that match the selectors.

    Returns:
        None

    """
    try:
        # Load manifests from local files.
        local_meta, _, err = manio.load(folder, selectors)
        assert not err and local_meta is not None

        # Download manifests from K8s.
        server, err = manio.download(config, client, selectors)
        assert not err and server is not None

        # Create the deployment plan.
        plan, err = compile_plan(config, local_meta, server)
        assert not err and plan is not None

        # Present the plan to the user.
        print_deltas(plan)

        # Ask for user confirmation. Abort if the user does not give it.
        if not user_confirmed(config.name):
            print("User abort - no changes were made.")
            return (None, True)

        # Create the missing resources.
        for data in plan.create:
            print(f"Creating {data.meta.kind.upper()} "
                  f"{data.meta.namespace}/{data.meta.name}")
            _, err = k8s.post(client, data.url, data.manifest)
            assert not err

        # Patch the server resources.
        patches = [_.patch for _ in plan.patch if len(_.patch.ops) > 0]
        print(f"Compiled {len(patches)} patches.")
        for patch in patches:
            pprint(patch)
            _, err = k8s.patch(client, patch.url, patch.ops)
            assert not err

        # Delete the excess resources.
        for data in plan.delete:
            print(f"Deleting {data.meta.kind.upper()} "
                  f"{data.meta.namespace}/{data.meta.name}")
            _, err = k8s.delete(client, data.url, data.manifest)
            assert not err
    except AssertionError:
        return (None, True)

    # All good.
    return (None, False)


def main_plan(
        config: Config,
        client,
        folder: Filepath,
        selectors: Selectors,
) -> Tuple[None, bool]:
    """Print the diff between `local_manifests` and `server_manifests`.

    The diff shows what would have to change on the K8s server in order for it
    to match the setup defined in `local_manifests`.

    Inputs:
        config: k8s.Config
        client: `requests` session with correct K8s certificates.
        folder: Path
            Path to local manifests eg "./foo"
        selectors: Selectors
            Only operate on resources that match the selectors.

    Returns:
        None

    """
    try:
        # Load manifests from local files.
        local_meta, _, err = manio.load(folder, selectors)
        assert not err and local_meta is not None

        # Download manifests from K8s.
        server, err = manio.download(config, client, selectors)
        assert not err and server is not None

        # Create deployment plan.
        plan, err = compile_plan(config, local_meta, server)
        assert not err and plan
    except AssertionError:
        return (None, True)

    # Print the plan and return.
    print_deltas(plan)
    return (None, False)


def main_get(
        config: Config,
        client,
        folder: Filepath,
        selectors: Selectors,
) -> Tuple[None, bool]:

    """Download all K8s manifests and merge them into local files.

    Inputs:
        config: k8s.Config
        client: `requests` session with correct K8s certificates.
        folder: Path
            Path to local manifests eg "./foo"
        selectors: Selectors
            Only operate on resources that match the selectors.

    Returns:
        None

    """
    try:
        # Load manifests from local files.
        _, local_files, err = manio.load(folder, selectors)
        assert not err and local_files is not None

        # Download manifests from K8s.
        server, err = manio.download(config, client, selectors)
        assert not err and server is not None

        # Sync the server manifests into the local manifests. All this happens in
        # memory and no files will be modified here - see `manio.save` in the next step.
        synced_manifests, err = manio.sync(local_files, server, selectors)
        assert not err and synced_manifests

        # Write the new manifest files.
        _, err = manio.save(folder, synced_manifests)
        assert not err
    except AssertionError:
        return (None, True)

    # Success.
    return (None, False)


def main() -> int:
    param = parse_commandline_args()

    if param.parser == "version":
        print(f"v{__version__}")
        return 0

    # Initialise logging.
    setup_logging(param.verbosity)

    # Create a `requests` client with proper security certificates to access the
    # Kubernetes API.
    kubeconfig = os.path.expanduser(param.kubeconfig)
    try:
        config = k8s.load_auto_config(kubeconfig, param.ctx, disable_warnings=True)
        assert config

        client = k8s.session(config)
        assert client

        # Update the config with the correct K8s API version.
        config, err = k8s.version(config, client)
        assert not err and config
    except AssertionError:
        return 1

    # Log the K8s API address and version.
    logit.info(f"Kubernetes server at {config.url}")
    logit.info(f"Kubernetes version is {config.version}")

    # Do what user asked us to do.
    selectors = Selectors(param.kinds, param.namespaces, set(param.labels))
    if param.parser == "get":
        _, err = main_get(config, client, param.folder, selectors)
    elif param.parser == "plan":
        _, err = main_plan(config, client, param.folder, selectors)
    elif param.parser == "apply":
        _, err = main_apply(config, client, param.folder, selectors)
    else:
        logit.error(f"Unknown command <{param.parser}>")
        return 1

    # Return error code.
    if err:
        return 1
    return 0


if __name__ == '__main__':
    sys.exit(main())



import pathlib
from collections import namedtuple
from typing import (
    Dict, Iterable, List, NamedTuple, Optional, Set, Tuple, Union,
)

# We support these resource types. The order matters because it determines the
# order in which the manifests will be grouped in the output files.
SUPPORTED_KINDS = (
    # Namespaces must come first to ensure the other resources can be created
    # within them.
    "Namespace",

    # Configuration and PVC before Deployments & friends use them.
    "ConfigMap", "Secret", "PersistentVolumeClaim",

    # RBAC.
    "ClusterRole", "ClusterRoleBinding", "Role", "RoleBinding", "ServiceAccount",

    # Define Services before creating Deployments & friends.
    "Service",

    # Everything that will spawn pods.
    "Deployment", "DaemonSet", "StatefulSet", "HorizontalPodAutoscaler",

    # Ingresses should be after Deployments & friends.
    "Ingress",
)


# Declare aliases for each resource type. Will be used in command line parsing
# to save the user some typing and match what `kubectl` would accept. We do not
# need to worry about capitalisation because `square.parse_commandline_args`
# will always convert everything to lower case letters first.
RESOURCE_ALIASES = {
    "ClusterRole": {"clusterrole", "clusterroles"},
    "ClusterRoleBinding": {"clusterrolebinding", "clusterrolebindings"},
    "ConfigMap": {"configmap", "cm"},
    "DaemonSet": {"daemonset", "daemonsets", "ds"},
    "Deployment": {"deployment", "deployments", "deploy"},
    "HorizontalPodAutoscaler": {"hpa"},
    "Ingress": {"ingress", "ingresses", "ing"},
    "Namespace": {"namespace", "namespaces", "ns"},
    "PersistentVolumeClaim": {"persistentVolumeClaim", "persistentvolumeclaims", "pvc"},
    "Role": {"role", "roles"},
    "RoleBinding": {"rolebinding", "rolebindings"},
    "Secret": {"secret", "secrets"},
    "Service": {"service", "services", "svc"},
    "ServiceAccount": {"serviceaccount", "serviceaccounts", "sa"},
    "StatefulSet": {"statefulset", "statefulsets"},
}

# Sanity check: we must have aliases for every supported resource kind.
assert set(SUPPORTED_KINDS) == set(RESOURCE_ALIASES.keys())

SUPPORTED_VERSIONS = ("1.9", "1.10", "1.11", "1.13", "1.14")

Config = namedtuple('Config', 'url token ca_cert client_cert version name')
DeltaCreate = namedtuple("DeltaCreate", "meta url manifest")
DeltaDelete = namedtuple("DeltaDelete", "meta url manifest")
DeltaPatch = namedtuple("Delta", "meta diff patch")
DeploymentPlan = namedtuple('DeploymentPlan', 'create patch delete')
JsonPatch = namedtuple('Patch', 'url ops')
MetaManifest = namedtuple('MetaManifest', 'apiVersion kind namespace name')

# Data types.
Filepath = Union[str, pathlib.PurePath]
LocalManifests = Dict[Filepath, Tuple[MetaManifest, dict]]
LocalManifestLists = Dict[Filepath, List[Tuple[MetaManifest, dict]]]
ServerManifests = Dict[MetaManifest, dict]


class Selectors(NamedTuple):
    """Comprises all the filters to select manifests."""
    kinds: Iterable[str]
    namespaces: Optional[Iterable[str]]
    labels: Optional[Set[Tuple[str, str]]]



"""Define the structure of the stripped manifest schema.

A stripped manifest is a sub-set of a normal manifest. The remaining keys
capture the salient information about the resource.

For instance, almost all manifest can have a "status" field. Albeit useful for
diagnostics, it makes no sense to compute diffs of "status" fields and submit
them in patches.

# Schema Conventions
Schemas are normal dictionaries without a depth limit.  All keys correspond to
a K8s manifest key. All value must be either dicts themselves, a bool to
specify whether the fields must be included in the stripped manifest, or None
if the field is not mandatory but should be included.

* True: field will be included. It is an error if the input manifest lacks it.
* False: field will not be included. It is an error if the input manifest has it.
* None: field will be included if the input manifest has it, and ignored otherwise.

"""
schema_1_9 = {
    "ClusterRole": {
        "metadata": {
            "labels": None,
            "name": True,
            "namespace": False,
        },
        "rules": None,
    },
    "ClusterRoleBinding": {
        "metadata": {
            "labels": None,
            "name": True,
            "namespace": False,
        },
        "roleRef": None,
        "subjects": None
    },
    "ConfigMap": {
        "metadata": {
            "labels": None,
            "name": True,
            "namespace": True,
        },
        "data": None,
        "binaryData": None,
    },
    "DaemonSet": {
        "metadata": {
            "labels": None,
            "name": True,
            "namespace": True,
        },
        "spec": True,
    },
    "Deployment": {
        "metadata": {
            "labels": None,
            "name": True,
            "namespace": True,
        },
        "creationTimestamp": False,
        "spec": True,
    },
    "HorizontalPodAutoscaler": {
        "metadata": {
            "labels": None,
            "name": True,
            "namespace": True,
        },
        "spec": True,
    },
    "Ingress": {
        "metadata": {
            "annotations": {
                "ingress.kubernetes.io/configuration-snippet": None,
                "kubernetes.io/ingress.class": None,
            },
            "labels": None,
            "name": True,
            "namespace": True,
        },
        "spec": True,
    },
    "Namespace": {
        "metadata": {
            "labels": None,
            "name": True,
            "namespace": False
        },
        "spec": None,
    },
    "PersistentVolumeClaim": {
        "metadata": {
            "labels": None,
            "name": True,
            "namespace": True,
        },
        "spec": True,
    },
    "Role": {
        "metadata": {
            "labels": None,
            "name": True,
            "namespace": True,
        },
        "rules": None,
    },
    "RoleBinding": {
        "metadata": {
            "labels": None,
            "name": True,
            "namespace": True,
        },
        "roleRef": None,
        "subjects": None
    },
    "Secret": {
        "metadata": {
            "labels": None,
            "name": True,
            "namespace": True,
            "annotations": {
                "kubernetes.io/service-account.name": None,
            }
        },
        "data": None,
        "stringData": None,
        "type": None,
    },
    "Service": {
        "metadata": {
            "labels": None,
            "name": True,
            "namespace": True,
        },
        "spec": {
            "ports": None,
            "selector": None,
            "type": True,
        },
    },
    "ServiceAccount": {
        "kind": None,
        "imagePullSecrets": None,
        "metadata": {
            "labels": None,
            "name": True,
            "namespace": True,
        },
        "secrets": None,
    },
    "StatefulSet": {
        "metadata": {
            "labels": None,
            "name": True,
            "namespace": True,
        },
        "spec": True,
    },
}


RESOURCE_SCHEMA = {
    "1.9": schema_1_9,
    "1.10": schema_1_9,
    "1.11": schema_1_9,
    "1.13": schema_1_9,
    "1.14": schema_1_9,
}



import base64
import json
import logging
import os
import pathlib
import re
import subprocess
import tempfile
import warnings
from collections import namedtuple
from typing import Optional, Tuple

import google.auth
import google.auth.transport.requests
import requests
import yaml
from square.dtypes import SUPPORTED_KINDS, SUPPORTED_VERSIONS, Config, Filepath

ClientCert = namedtuple('ClientCert', 'crt key')

FNAME_TOKEN = "/var/run/secrets/kubernetes.io/serviceaccount/token"
FNAME_CERT = "/var/run/secrets/kubernetes.io/serviceaccount/ca.crt"


# Convenience: global logger instance to avoid repetitive code.
logit = logging.getLogger("square")


def load_kubeconfig(
        fname: Filepath,
        context: Optional[str]
) -> Tuple[Optional[str], Optional[dict], Optional[dict]]:
    """Return user name and user- and cluster information.

    Return None on error.

    Inputs:
        fname: str
            Path to kubeconfig file, eg "~/.kube/config.yaml"
        context: str
            Kubeconf context. Use `None` to use default context.

    Returns:
        name, user info, cluster info

    """
    # Load `kubeconfig`.
    try:
        kubeconf = yaml.safe_load(open(fname))
    except (IOError, PermissionError) as err:
        logit.error(f"{err}")
        return (None, None, None)

    # Find the user and cluster information based on the specified `context`.
    try:
        # Use default context unless specified.
        ctx = context if context else kubeconf["current-context"]
        del context

        try:
            # Find the correct context.
            ctx = [_ for _ in kubeconf["contexts"] if _["name"] == ctx]
            assert len(ctx) == 1
            ctx = ctx[0]["context"]

            # Unpack the cluster- and user name from the current context.
            clustername, username = ctx["cluster"], ctx["user"]

            # Find the information for the current cluster and user.
            user_info = [_ for _ in kubeconf["users"] if _["name"] == username]
            cluster_info = [_ for _ in kubeconf["clusters"] if _["name"] == clustername]
            assert len(user_info) == len(cluster_info) == 1
        except AssertionError:
            logit.error(f"Could not find information for context <{ctx}>")
            return (None, None, None)

        # Unpack the cluster- and user information.
        cluster_name = cluster_info[0]["name"]
        cluster_info_out = cluster_info[0]["cluster"]
        cluster_info_out["name"] = cluster_name
        user_info = user_info[0]["user"]
        del cluster_info
    except (KeyError, TypeError):
        logit.error(f"Kubeconfig YAML file <{fname}> is invalid")
        return (None, None, None)

    # Success. The explicit `dicts()` are to satisfy MyPy.
    logit.info(f"Loaded {ctx} from Kubeconfig file <{fname}>")
    return (username, dict(user_info), dict(cluster_info_out))


def load_incluster_config(
        fname_token: Filepath = FNAME_TOKEN,
        fname_cert: Filepath = FNAME_CERT) -> Optional[Config]:
    """Return K8s access config from Pod service account.

    Returns None if we are not running in a Pod.

    Inputs:
        kubconfig: str
            Name of kubeconfig file.
    Returns:
        Config

    """
    # Every K8s pod has this.
    server_ip = os.getenv('KUBERNETES_PORT_443_TCP_ADDR', None)

    fname_cert = pathlib.Path(fname_cert)
    fname_token = pathlib.Path(fname_token)

    # Sanity checks: URL and service account files either exist, or we are not
    # actually inside a Pod.
    try:
        assert server_ip is not None
        assert fname_cert.exists()
        assert fname_token.exists()
    except AssertionError:
        logit.debug("Could not find incluster (service account) credentials.")
        return None

    # Return the compiled K8s access configuration.
    logit.info("Use incluster (service account) credentials.")
    return Config(
        url=f'https://{server_ip}',
        token=fname_token.read_text(),
        ca_cert=str(fname_cert),
        client_cert=None,
        version=None,
        name="",
    )


def load_gke_config(
        fname: Filepath,
        context: Optional[str],
        disable_warnings: bool = False) -> Optional[Config]:
    """Return K8s access config for GKE cluster described in `kubeconfig`.

    Returns None if `kubeconfig` does not exist or could not be parsed.

    Inputs:
        kubconfig: str
            Name of kubeconfig file.
        context: str
            Kubeconf context. Use `None` to use default context.
        disable_warnings: bool
            Whether or not do disable GCloud warnings.

    Returns:
        Config

    """
    # Parse the kubeconfig file.
    name, user, cluster = load_kubeconfig(fname, context)
    if name is None or user is None or cluster is None:
        return None

    # Unpack the self signed certificate (Google does not register the K8s API
    # server certificate with a public CA).
    try:
        ssl_ca_cert_data = base64.b64decode(
            cluster["certificate-authority-data"]
        )
    except KeyError:
        logit.debug(f"Context {context} in <{fname}> is not a GKE config")
        return None

    # Save the certificate to a temporary file. This is only necessary because
    # the requests library will need a path to the CA file - unfortunately, we
    # cannot just pass it the content.
    _, ssl_ca_cert = tempfile.mkstemp(text=False)
    with open(ssl_ca_cert, "wb") as fd:
        fd.write(ssl_ca_cert_data)

    with warnings.catch_warnings(record=disable_warnings):
        cred, project_id = google.auth.default(
            scopes=['https://www.googleapis.com/auth/cloud-platform']
        )
        cred.refresh(google.auth.transport.requests.Request())
        token = cred.token

    # Return the config data.
    logit.info(f"Assuming GKE cluster.")
    return Config(
        url=cluster["server"],
        token=token,
        ca_cert=ssl_ca_cert,
        client_cert=None,
        version=None,
        name=cluster["name"],
    )


def load_eks_config(
        fname: Filepath,
        context: Optional[str],
        disable_warnings: bool = False) -> Optional[Config]:
    """Return K8s access config for EKS cluster described in `kubeconfig`.

    Returns None if `kubeconfig` does not exist or could not be parsed.

    Inputs:
        kubconfig: str
            Name of kubeconfig file.
        context: str
            Kubeconf context. Use `None` to use default context.
        disable_warnings: bool
            Whether or not do disable GCloud warnings.

    Returns:
        Config

    """
    # Parse the kubeconfig file.
    name, user, cluster = load_kubeconfig(fname, context)
    if name is None or user is None or cluster is None:
        return None

    # Get a copy of all env vars. We will pass that one along to the
    # sub-process, plus the env vars specified in the kubeconfig file.
    env = os.environ.copy()

    # Unpack the self signed certificate (AWS does not register the K8s API
    # server certificate with a public CA).
    try:
        ssl_ca_cert_data = base64.b64decode(cluster["certificate-authority-data"])
        cmd = user["exec"]["command"]
        args = user["exec"]["args"]
        env_kubeconf = user["exec"].get("env", [])
    except KeyError:
        logit.debug(f"Context {context} in <{fname}> is not an EKS config")
        return None

    # Convert a None value (valid value in YAML) to an empty list of env vars.
    env_kubeconf = env_kubeconf if env_kubeconf is not None else []

    # Save the certificate to a temporary file. This is only necessary because
    # the requests library will need a path to the CA file - unfortunately, we
    # cannot just pass it the content.
    _, ssl_ca_cert = tempfile.mkstemp(text=False)
    with open(ssl_ca_cert, "wb") as fd:
        fd.write(ssl_ca_cert_data)

    # Compile the name, arguments and env vars for the command specified in kubeconf.
    cmd_args = [cmd] + args
    env_kubeconf = {_["name"]: _["value"] for _ in env_kubeconf}
    env.update(env_kubeconf)
    logit.debug(f"Requesting EKS certificate: {cmd_args} with envs: {env_kubeconf}")

    # Run the specified command to produce the access token. That program must
    # produce a YAML document on stdout that specifies the bearer token.
    try:
        out = subprocess.run(cmd_args, stdout=subprocess.PIPE, env=env)
        token = yaml.safe_load(out.stdout.decode("utf8"))["status"]["token"]
    except FileNotFoundError:
        logit.error(f"Could not find <{cmd}> application to get token")
        return None
    except (KeyError, yaml.YAMLError):
        logit.error(f"Token manifest from <{cmd}> is corrupt")
        return None

    # Return the config data.
    logit.info(f"Assuming EKS cluster.")
    return Config(
        url=cluster["server"],
        token=token,
        ca_cert=ssl_ca_cert,
        client_cert=None,
        version=None,
        name=cluster["name"],
    )


def load_minikube_config(
        fname: Filepath,
        context: Optional[str],
) -> Optional[Config]:
    """Load minikube configuration from `fname`.

    Return None on error.

    Inputs:
        kubconfig: str
            Path to kubeconfig file, eg "~/.kube/config.yaml"
        context: str
            Kubeconf context. Use `None` to use default context.

    Returns:
        Config

    """
    # Parse the kubeconfig file.
    name, user, cluster = load_kubeconfig(fname, context)
    if name is None or user is None or cluster is None:
        return None

    # Minikube uses client certificates to authenticate. We need to pass those
    # to the HTTP client of our choice when we create the session.
    try:
        client_cert = ClientCert(
            crt=user["client-certificate"],
            key=user["client-key"],
        )

        # Return the config data.
        logit.info(f"Assuming Minikube cluster.")
        return Config(
            url=cluster["server"],
            token=None,
            ca_cert=cluster["certificate-authority"],
            client_cert=client_cert,
            version=None,
            name=cluster["name"],
        )
    except KeyError:
        logit.debug(f"Context {context} in <{fname}> is not a Minikube config")
        return None


def load_kind_config(
        fname: Filepath,
        context: Optional[str],
) -> Optional[Config]:
    """Load Kind configuration from `fname`.

    https://github.com/bsycorp/kind

    Kind is just another Minikube cluster. The only notable difference
    is that it does not store its credentials as files but directly in
    the Kubeconfig file. This function will copy those files into /tmp.

    Return None on error.

    Inputs:
        kubconfig: str
            Path to kubeconfig for Kind cluster.
        context: str
            Kubeconf context. Use `None` to use default context.

    Returns:
        Config

    """
    # Parse the kubeconfig file.
    name, user, cluster = load_kubeconfig(fname, context)
    if name is None or user is None or cluster is None:
        return None

    # Kind/Minikube use client certificates to authenticate. We need to pass
    # those to the HTTP client of our choice when we create the session.
    try:
        client_crt = base64.b64decode(user["client-certificate-data"]).decode()
        client_key = base64.b64decode(user["client-key-data"]).decode()
        client_ca = base64.b64decode(cluster["certificate-authority-data"]).decode()
        path = pathlib.Path("/tmp/")  # nosec
        p_client_crt = path / "kind-client.crt"
        p_client_key = path / "kind-client.key"
        p_ca = path / "kind.ca"
        open(p_client_crt, "w").write(client_crt)
        open(p_client_key, "w").write(client_key)
        open(p_ca, "w").write(client_ca)
        client_cert = ClientCert(crt=p_client_crt, key=p_client_key)

        # Return the config data.
        logit.debug(f"Assuming Minikube/Kind cluster.")
        return Config(
            url=cluster["server"],
            token=None,
            ca_cert=p_ca,
            client_cert=client_cert,
            version=None,
            name=cluster["name"],
        )
    except KeyError:
        logit.debug(f"Context {context} in <{fname}> is not a Minikube config")
        return None


def load_auto_config(
        fname: Filepath,
        context: Optional[str],
        disable_warnings: bool = False) -> Optional[Config]:
    """Automagically find and load the correct K8s configuration.

    This function will load several possible configuration options and returns
    the first one with a match. The order is as follows:

    1) `load_incluster_config`
    2) `load_gke_config`

    Inputs:
        fname: str
            Path to kubeconfig file, eg "~/.kube/config.yaml"
            Use `None` to find out automatically or for incluster credentials.

        context: str
            Kubeconf context. Use `None` to use default context.

    Returns:
        Config

    """
    conf = load_incluster_config()
    if conf is not None:
        return conf
    logit.debug("Incluster config failed")

    conf = load_minikube_config(fname, context)
    if conf is not None:
        return conf
    logit.debug("Minikube config failed")

    conf = load_kind_config(fname, context)
    if conf is not None:
        return conf
    logit.debug("KIND config failed")

    conf = load_eks_config(fname, context, disable_warnings)
    if conf is not None:
        return conf
    logit.debug("EKS config failed")

    conf = load_gke_config(fname, context, disable_warnings)
    if conf is not None:
        return conf
    logit.debug("GKE config failed")

    logit.error(f"Could not find a valid configuration in <{fname}>")
    return None


def session(config: Config):
    """Return configured `requests` session."""
    # Plain session.
    sess = requests.Session()

    # Load the CA file (necessary for self signed certs to avoid https warning).
    sess.verify = config.ca_cert

    # Add the client certificate, if the cluster uses those to authenticate users.
    if config.client_cert is not None:
        sess.cert = (config.client_cert.crt, config.client_cert.key)

    # Add the bearer token if this cluster uses them to authenticate users.
    if config.token is not None:
        sess.headers.update({'authorization': f'Bearer {config.token}'})

    # Return the configured session object.
    return sess


def urlpath(
        config: Config,
        kind: str,
        namespace: Optional[str]) -> Tuple[Optional[str], bool]:
    """Return complete URL to K8s resource.

    Inputs:
        config: k8s.Config
        kind: str
            Eg "Deployment", "Namespace", ... (case sensitive).
        namespace: str
            Must be None for "Namespace" resources.

    Returns:
        str: Full path K8s resource, eg https://1.2.3.4/api/v1/namespace/foo/services.

    """
    # Namespaces are special because they lack the `namespaces/` path prefix.
    if kind == "Namespace" or namespace is None:
        namespace = ""
    else:
        # Namespace name must conform to K8s standards.
        match = re.match(r"[a-z0-9]([-a-z0-9]*[a-z0-9])?", namespace)
        if match is None or match.group() != namespace:
            logit.error(f"Invalid namespace name <{namespace}>.")
            return (None, True)

        namespace = f"namespaces/{namespace}/"

    # We must support the specified resource kind.
    if kind not in SUPPORTED_KINDS:
        logit.error(f"Unsupported resource <{kind}>.")
        return (None, True)

    # We must support the specified K8s version.
    if config.version not in SUPPORTED_VERSIONS:
        logit.error(f"Unsupported K8s version <{config.version}>.")
        return (None, True)

    # The HTTP request path names by K8s version and resource kind.
    # The keys in this dict must cover all those specified in
    # `SUPPORTED_VERSIONS` and `SUPPORTED_KINDS`.
    resources = {
        "1.9": {
            "ClusterRole": f"apis/rbac.authorization.k8s.io/v1/clusterroles",
            "ClusterRoleBinding": f"apis/rbac.authorization.k8s.io/v1/clusterrolebindings",  # NOQA
            "ConfigMap": f"api/v1/{namespace}/configmaps",
            "DaemonSet": f"apis/apps/v1/{namespace}/daemonsets",
            "Deployment": f"apis/extensions/v1beta1/{namespace}/deployments",
            "HorizontalPodAutoscaler": f"apis/autoscaling/v1/{namespace}/horizontalpodautoscalers",   # NOQA
            "Ingress": f"apis/extensions/v1beta1/{namespace}/ingresses",
            "Namespace": f"api/v1/namespaces",
            "PersistentVolumeClaim": f"api/v1/{namespace}/persistentvolumeclaims",
            "Role": f"apis/rbac.authorization.k8s.io/v1/{namespace}/roles",
            "RoleBinding": f"apis/rbac.authorization.k8s.io/v1/{namespace}/rolebindings",
            "Secret": f"api/v1/{namespace}/secrets",
            "Service": f"api/v1/{namespace}/services",
            "ServiceAccount": f"api/v1/{namespace}/serviceaccounts",
            "StatefulSet": f"apis/apps/v1/{namespace}/statefulsets",
        },
        "1.10": {
            "ClusterRole": f"apis/rbac.authorization.k8s.io/v1/clusterroles",
            "ClusterRoleBinding": f"apis/rbac.authorization.k8s.io/v1/clusterrolebindings",  # NOQA
            "ConfigMap": f"api/v1/{namespace}/configmaps",
            "DaemonSet": f"apis/apps/v1/{namespace}/daemonsets",
            "Deployment": f"apis/extensions/v1beta1/{namespace}/deployments",
            "HorizontalPodAutoscaler": f"apis/autoscaling/v1/{namespace}/horizontalpodautoscalers",   # NOQA
            "Ingress": f"apis/extensions/v1beta1/{namespace}/ingresses",
            "Namespace": f"api/v1/namespaces",
            "PersistentVolumeClaim": f"api/v1/{namespace}/persistentvolumeclaims",
            "Role": f"apis/rbac.authorization.k8s.io/v1/{namespace}/roles",
            "RoleBinding": f"apis/rbac.authorization.k8s.io/v1/{namespace}/rolebindings",
            "Secret": f"api/v1/{namespace}/secrets",
            "Service": f"api/v1/{namespace}/services",
            "ServiceAccount": f"api/v1/{namespace}/serviceaccounts",
            "StatefulSet": f"apis/apps/v1/{namespace}/statefulsets",
        },
        "1.11": {
            "ClusterRole": f"apis/rbac.authorization.k8s.io/v1/clusterroles",
            "ClusterRoleBinding": f"apis/rbac.authorization.k8s.io/v1/clusterrolebindings",  # NOQA
            "ConfigMap": f"api/v1/{namespace}/configmaps",
            "DaemonSet": f"apis/apps/v1/{namespace}/daemonsets",
            "Deployment": f"apis/apps/v1/{namespace}/deployments",
            "HorizontalPodAutoscaler": f"apis/autoscaling/v1/{namespace}/horizontalpodautoscalers",   # NOQA
            "Ingress": f"apis/extensions/v1beta1/{namespace}/ingresses",
            "Namespace": f"api/v1/namespaces",
            "PersistentVolumeClaim": f"api/v1/{namespace}/persistentvolumeclaims",
            "Role": f"apis/rbac.authorization.k8s.io/v1/{namespace}/roles",
            "RoleBinding": f"apis/rbac.authorization.k8s.io/v1/{namespace}/rolebindings",
            "Secret": f"api/v1/{namespace}/secrets",
            "Service": f"api/v1/{namespace}/services",
            "ServiceAccount": f"api/v1/{namespace}/serviceaccounts",
            "StatefulSet": f"apis/apps/v1/{namespace}/statefulsets",
        },
        "1.13": {
            "ClusterRole": f"apis/rbac.authorization.k8s.io/v1/clusterroles",
            "ClusterRoleBinding": f"apis/rbac.authorization.k8s.io/v1/clusterrolebindings",  # NOQA
            "ConfigMap": f"api/v1/{namespace}/configmaps",
            "DaemonSet": f"apis/apps/v1/{namespace}/daemonsets",
            "Deployment": f"apis/apps/v1/{namespace}/deployments",
            "HorizontalPodAutoscaler": f"apis/autoscaling/v1/{namespace}/horizontalpodautoscalers",   # NOQA
            "Ingress": f"apis/extensions/v1beta1/{namespace}/ingresses",
            "Namespace": f"api/v1/namespaces",
            "PersistentVolumeClaim": f"api/v1/{namespace}/persistentvolumeclaims",
            "Role": f"apis/rbac.authorization.k8s.io/v1/{namespace}/roles",
            "RoleBinding": f"apis/rbac.authorization.k8s.io/v1/{namespace}/rolebindings",
            "Secret": f"api/v1/{namespace}/secrets",
            "Service": f"api/v1/{namespace}/services",
            "ServiceAccount": f"api/v1/{namespace}/serviceaccounts",
            "StatefulSet": f"apis/apps/v1/{namespace}/statefulsets",
        },
        "1.14": {
            "ClusterRole": f"apis/rbac.authorization.k8s.io/v1/clusterroles",
            "ClusterRoleBinding": f"apis/rbac.authorization.k8s.io/v1/clusterrolebindings",  # NOQA
            "ConfigMap": f"api/v1/{namespace}/configmaps",
            "DaemonSet": f"apis/apps/v1/{namespace}/daemonsets",
            "Deployment": f"apis/apps/v1/{namespace}/deployments",
            "HorizontalPodAutoscaler": f"apis/autoscaling/v1/{namespace}/horizontalpodautoscalers",   # NOQA
            "Ingress": f"apis/extensions/v1beta1/{namespace}/ingresses",
            "Namespace": f"api/v1/namespaces",
            "PersistentVolumeClaim": f"api/v1/{namespace}/persistentvolumeclaims",
            "Role": f"apis/rbac.authorization.k8s.io/v1/{namespace}/roles",
            "RoleBinding": f"apis/rbac.authorization.k8s.io/v1/{namespace}/rolebindings",
            "Secret": f"api/v1/{namespace}/secrets",
            "Service": f"api/v1/{namespace}/services",
            "ServiceAccount": f"api/v1/{namespace}/serviceaccounts",
            "StatefulSet": f"apis/apps/v1/{namespace}/statefulsets",
        },
    }

    # Look up the resource path and remove duplicate "/" characters (may have
    # slipped in when no namespace was supplied, eg "/api/v1//configmaps").
    path = resources[config.version][kind]
    path = path.replace("//", "/")
    assert not path.startswith("/")

    # Return the complete resource URL.
    return (f"{config.url}/{path}", False)


def request(
        client,
        method: str,
        url: str,
        payload: Optional[dict],
        headers: Optional[dict]) -> Tuple[Optional[dict], bool]:
    """Return response of web request made with `client`.

    Inputs:
        client: `requests` session with correct K8s certificates.
        url: str
            Eg `https://1.2.3.4/api/v1/namespaces`)
        payload: dict
            Anything that can be JSON encoded, usually a K8s manifest.
        headers: dict
            Request headers. These will *not* replace the existing request
            headers dictionary (eg the access tokens), but augment them.

    Returns:
        (dict, int): the JSON response and the HTTP status code.

    """
    try:
        ret = client.request(method, url, json=payload, headers=headers, timeout=30)
    except requests.exceptions.ConnectionError as err:
        method = err.request.method
        url = err.request.url
        logit.error(f"Connection error: {method} {url}")
        return (None, True)

    try:
        response = ret.json()
    except json.decoder.JSONDecodeError as err:
        msg = (
            f"JSON error: {err.msg} in line {err.lineno} column {err.colno}",
            "-" * 80 + "\n" + err.doc + "\n" + "-" * 80,
        )
        logit.error(str.join("\n", msg))
        return (None, True)

    # Log the entire request in debug mode.
    logit.debug(
        f"{method} {ret.status_code} {ret.url}\n"
        f"Headers: {headers}\n"
        f"Payload: {payload}\n"
        f"Response: {response}\n"
    )
    return (response, ret.status_code)


def delete(client, url: str, payload: dict) -> Tuple[Optional[dict], bool]:
    """Make DELETE requests to K8s (see `k8s_request`)."""
    resp, code = request(client, 'DELETE', url, payload, headers=None)
    err = (code not in (200, 202))
    if err:
        logit.error(f"{code} - DELETE - {url} - {resp}")
    return (resp, err)


def get(client, url: str) -> Tuple[Optional[dict], bool]:
    """Make GET requests to K8s (see `request`)."""
    resp, code = request(client, 'GET', url, payload=None, headers=None)
    err = (code != 200)
    if err:
        logit.error(f"{code} - GET - {url} - {resp}")
    return (resp, err)


def patch(client, url: str, payload: dict) -> Tuple[Optional[dict], bool]:
    """Make PATCH requests to K8s (see `request`)."""
    headers = {'Content-Type': 'application/json-patch+json'}
    resp, code = request(client, 'PATCH', url, payload, headers)
    err = (code != 200)
    if err:
        logit.error(f"{code} - PATCH - {url} - {resp}")
    return (resp, err)


def post(client, url: str, payload: dict) -> Tuple[Optional[dict], bool]:
    """Make POST requests to K8s (see `request`)."""
    resp, code = request(client, 'POST', url, payload, headers=None)
    err = (code != 201)
    if err:
        logit.error(f"{code} - POST - {url} - {resp}")
    return (resp, err)


def version(config: Config, client) -> Tuple[Optional[Config], bool]:
    """Return new `config` with version number of K8s API.

    Contact the K8s API, query its version via `client` and return `config`
    with an updated `version` field. All other field in `config` will remain
    intact.

    Inputs:
        config: Config
        client: `requests` session with correct K8s certificates.

    Returns:
        Config

    """
    # Ask the K8s API for its version and check for errors.
    url = f"{config.url}/version"
    resp, err = get(client, url)
    if err or resp is None:
        return (None, True)

    # Construct the version number of the K8s API.
    major, minor = resp['major'], resp['minor']
    version = f"{major}.{minor}"

    # If we are talking to GKE, the version string may now be "1.10+". It
    # simply indicates that GKE is running version 1.10.x. We need to remove
    # the "+" because the version string is important in `square`, for instance
    # to determines which URLs to contact, which fields are valid.
    version = version.replace("+", "")

    # Return an updated `Config` tuple.
    config = config._replace(version=version)
    return (config, False)



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['square']

package_data = \
{'': ['*']}

install_requires = \
['colorama>=0.4.1,<0.5.0',
 'google-api-python-client>=1.7,<2.0',
 'google>=2.0,<3.0',
 'jsonpatch>=1.24,<2.0',
 'pyyaml>=5.1,<6.0',
 'requests>=2.22,<3.0']

entry_points = \
{'console_scripts': ['square = square.square:main']}

setup_kwargs = {
    'name': 'kubernetes-square',
    'version': '0.13.0',
    'description': '',
    'long_description': None,
    'author': 'Oliver Nagy',
    'author_email': 'olitheolix@gmail.com',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
